#include	"u.h"
#include	"../port/lib.h"
#include	"mem.h"
#include	"dat.h"
#include	"fns.h"
#include	"io.h"

/*
 * system clock using the UltraSparc's tick compare register 
 */

enum {
	Fudge = 300,					/* ensure next tick is at least this far ahead */
};

static ulong clockfreq;
static ulong cyclespertick;

static void
clockreset(void)
{
	uvlong tick, nexttick;

	intrdisable();
	tick = gettick();
	nexttick = ((tick + cyclespertick - 1) / cyclespertick) * cyclespertick;
	if(nexttick - tick < Fudge)
		nexttick += cyclespertick;
	puttickcmp(nexttick);
	intrenable();
}

void
clockinit(void)
{
	/* not sure where the factor of 2 comes from -- but thats what measurements indicate */
	clockfreq = conf.cpufreq / 2;
	cyclespertick = clockfreq / HZ;
	clockreset();
}

void
clockintr(Ureg *ur)
{
	timerintr(ur, 0);
	clockreset();
}

/* port: return a fast running clock and its frequency */
uvlong
fastticks(uvlong* hz)
{
	if(hz)
		*hz = clockfreq;
	return gettick();
}

/* port: return a cycle count */
void
cycles(uvlong *cy)
{
	*cy = gettick();
}

/* port */
ulong
perfticks(void)
{
	return gettick();
}

void
microdelay(int us)
{
	uvlong start, eot;

	start = gettick();
	eot = start + us * (uvlong)clockfreq / 1000000LL;
	while(gettick() < eot)
		continue;
}

/* port */
void
delay(int ms)
{
	microdelay(ms * 1000);
}

/* port */
void
timerset(uvlong next)
{
	/* XXX we don't provide anything more fine grained right now */
	USED(next);
}
