/*
 * machine specific data definitions.
 * All data structures and fields required by the port code is marked "port".
 */

typedef struct Conf		Conf;		/* port */
typedef struct FPsave	FPsave;		/* port */
typedef struct Label		Label;		/* port */
typedef struct Lock		Lock;			/* port */
typedef struct Mach		Mach;		/* port */
typedef struct Notsave	Notsave;		/* port */
typedef struct PMMU	PMMU;		/* port */
typedef struct Proc		Proc;			/* port */
typedef struct Ureg		Ureg;		/* port */

typedef void			KMap;		/* port */

#define MAXSYSARG		5	/* port: for mount(fd, afp, mpt, flag, arg) */
#define MAXMACH		1	/* port */

/*
 *  port: parameters for sysproc.c
 */
#define AOUT_MAGIC	U_MAGIC

/*
 *  port: machine dependent definitions used by ../port/dat.h
 */
struct Lock
{
	ulong	key;		/* port */
	ulong	sr;		/* port */
	ulong	pc;		/* port */
	Proc		*p;		/* port */
	ushort	isilock;	/* port */
};

/* port */
struct Label
{
	ulong sp;
	ulong pc;
};

/*
 * floating point registers
 */
enum
{
	FPinit,		/* port */
	FPactive,
	FPinactive,
	FPillegal = 0x100
};

/* port */
struct	FPsave
{
	long fpreg[32];
	long fsr;
	uchar fprs;
};

#define NPROPS	32

/* port */
struct Conf
{
	ulong	nmach;		/* port: processors */
	ulong	nproc;		/* port: processes */
	ulong	monitor;		/* port: graphics monitor id; 0 for none */
	ulong	npage0;		/* port: total physical pages of memory, bank 0 */
	ulong	npage1;		/* port: total physical pages of memory, bank 1 */
	ulong	base0;		/* port: base of bank 0 */
	ulong	base1;		/* port: base of bank 1 */
	ulong	npage;		/* port: total physical pages of memory */
	ulong	nimage;		/* port: number of page cache image headers */
	ulong	nswap;		/* port: number of swap blocks */
	ulong	upages;		/* port: number of user pages */
	int		copymode;	/* port: 0 is copy on write, 1 is copy on reference */
	ulong	ialloc;		/* port: bytes available for interrupt time allocation */
	ulong	pipeqsize;		/* port: size in bytes of pipe queues */
	int		nswppo;		/* port: max # of pageouts per segment pass */

	/* port-specific */
	int		nprops;
	char*	propname[NPROPS];
	char*	propvalue[NPROPS];
	uchar	idprom[16];
	ulong	cpufreq;		/* change when we hit 4GHz :) */
};


/*
 *  mmu goo in the Proc structure
 */
/* port */
typedef struct Page Page;
struct PMMU
{
	int		ctx;			/* associated context */
};

/*
 *  port: things saved in the Proc structure during a notify
 */
struct Notsave
{
	ulong	XXXdummy;
};

#include "../port/portdat.h"

/* port */
struct Mach
{
	/* OFFSETS OF THE FOLLOWING KNOWN BY l.s */
	int		machno;		/* 0: port: physical id of processor */
	ulong	splpc;		/* 4: port: pc of last caller to splhi */
	Proc		*proc;		/* 8: port: current process on this processor */

	/* ordering from here on irrelevant */
	ulong	ticks;		/* port: of the clock since boot time */
	Label	sched;		/* port: scheduler wakeup */
	int		tlbfault;		/* port */
	int		tlbpurge;		/* port */
	int		pfault;		/* port */
	int		cs;			/* port */
	int		syscall;		/* port */
	int		load;			/* port */
	int		intr;			/* port */
	Perf		perf;			/* port: performance counters */
	int		flushmmu;	/* port: make current proc flush it's mmu state */
	int		ilockdepth;	/* port */
	Proc*	readied;		/* port: for runproc */
	ulong	schedticks;	/* port: next forced context switch */
	uvlong	cyclefreq;		/* port: Frequency of user readable cycle counter */

	/* port-specific fields */
	ulong	stack;
};




extern Mach	mach0;
extern register Mach	*m;		/* port: R6 */
extern register Proc	*up;		/* R5 */

#define	MACHP(n)	(n==0? &mach0 : *(Mach**)0)	/*port */


/* port */
struct
{
	Lock;
	short		machs;			/* port */
	short		exiting;			/* port */
	short 	ispanic;
} active;

