
#include "instr64.h"
#include "mem.h"

/*
 * N.B. R14 is often used as a temporary by the assembler
 * for instructions like MOVW of a large immediate.
 */

#define UREGSZ	(4*(32+6))
#define BIAS		(2048-1)
#define FRSIZE		176

GLOBL	ofwentry(SB), $4
GLOBL	fsr(SB), $4

/*
 * RMO is the best memory order, but lets wait until we start
 * debugging SMP before we turn it on.
 */
#define MEMORDER	PSTSO
#define PSKERNEL	(PSPEF|PSAM|PSPRIV|MEMORDER)
#define PSUSER	(PSAM|PSIE|MEMORDER)
#define CRASH		NOSCHED; WORD $1; SCHED

/*
 * Entry point from openfirmware. 
 * We're called with the client interface handle in %o4
 * and our stack pointer in %o6.  
 * We're in low memory with physical addresses equal
 * to virtual addresses.
 * Since we're 64-bits, %o6 is offset by BIAS.
 */
TEXT start(SB),$-4

	/* map phys 0..4M to virt KZERO in itlb and dtlb */
	MOVW	$MMU_TAGACCESS, R8
	MOVW	$KZERO, R9
	STXAr(9, 8, 0, ASI_MMU_I)
	STXAr(9, 8, 0, ASI_MMU_D)
	MEMBAR(MBSYNC)
	MOVW	$0xe0000000, R8
	SLLXi(8, 8, 32)
	OR		$(PTELOCK|PTEKERNEL|PTEWRITE|PTEGLOB | 0x30), R8
	MOVW	$(63<<3), R9
	STXAr(8, 9, 0, ASI_MMU_IDATA)
	STXAr(8, 9, 0, ASI_MMU_DDATA)
	MEMBAR(MBSYNC)

	MOVW	$startvirt(SB), R7
	JMPL		(R7)

TEXT startvirt(SB), $-4
	/* setup R1, R2 and ofwentry */
	MOVW	$setSB(SB), R2
	MOVW	$(BOOTSTACK+KSTACK), R1
	MOVW	R12, ofwentry(SB)

	WRPRi(RPIL, 0, 15)
	WRPRi(RPSTATE, 0, PSKERNEL)
	WRPRi(RCWP, 0, 0)
	WRSTi(RASI, 0, 0);
	WRPRi(RTICK, 0, 0)

	/* XXX same flags as the old sparc port used.  revisit this */
	MOVW	$(0x35 << 22), R8	/* NVM OFM DZM NS */
	MOVW	R8, fsr(SB)
	MOVW	$fsr(SB), R8
	MOVW	(R8), FSR
	WRSTi(RFPRS, 0, 0x4)		/* FPRS with fp enabled */
	FMOVD	$0.5, F26			/* 0.5 -> F26 */
	FSUBD	F26, F26, F24		/* 0.0 -> F24 */
	FADDD	F26, F26, F28		/* 1.0 -> F28 */
	FADDD	F28, F28, F30		/* 2.0 -> F30 */
	FMOVD	F24, F0
	FMOVD	F24, F2
	FMOVD	F24, F4
	FMOVD	F24, F6
	FMOVD	F24, F8
	FMOVD	F24, F10
	FMOVD	F24, F12
	FMOVD	F24, F14
	FMOVD	F24, F16
	FMOVD	F24, F18
	FMOVD	F24, F20
	FMOVD	F24, F22

	JMPL		main(SB)

TEXT getasi4(SB), $0				/* uchar ulong -> ulong */
	WRSTi(RASI, 7, 0)
	MOVW	8(R1), R8
	LDWAi(7, 8, 0)
	RETURN

TEXT getasi8(SB), $0				/* uchar ulong -> uvlong */
	MOVW	8(R1), R8
	WRSTi(RASI, 8, 0)
	MOVW	12(R1), R8
	LDXAi(9, 8, 0)

	// XXX alignment! STXi(9, 7, 0)
	MOVW	R9, 4(R7)
	SRLXi(9, 9, 32)
	MOVW	R9, 0(R7)
	RETURN

#ifdef notneeded
TEXT putasi4(SB), $0			/* uchar ulong ulong -> void */
	WRSTi(RASI, 7, 0)
	MOVW	8(R1), R8
	MOVW	12(R1), R9
	STWAi(9, 8, 0)
	RETURN
#endif

TEXT putasi8(SB), $0			/* uchar ulong uvlong -> void */
	WRSTi(RASI, 7, 0)
	MOVW	8(R1), R8
	MOVW	12(R1), R9
	MOVW	16(R1), R10
	SLLXi(9, 9, 32)
	OR		R10, R9

	/* interrupts off so MBSYNC happens immediately after store */
	RDPR(10, RPSTATE)
	WRPRi(RPSTATE, 0, PSKERNEL)
	STXAi(9, 8, 0)
	MEMBAR(MBSYNC)			/* ultrasparc I user's manual 5.3.10 */
	WRPRi(RPSTATE, 10, 0)
	/* XXX flush for mmu ops? */
	RETURN

/* could be easier if stack was aligned */
TEXT gettick(SB), $0				/* void -> uvlong */
	RDST(9, RTICK)
	MOVW	R9, 4(R7)
	SRLXi(9, 9, 32)
	MOVW	R9, 0(R7)
	RETURN

TEXT puttickcmp(SB), $0			/* uvlong -> void */
	MOVW	4(R1), R7
	SLLXi(7, 7, 32)
	MOVW	8(R1), R8
	OR		R8, R7
	WRSTi(RTICKCMP, 7, 0)
	RETURN
	

/* atomic 8-bit IO (needed for io registers) */
TEXT put8(SB), $0				/* ulong uvlong -> void */
	MOVW	8(R1), R8
	SLLXi(8, 8, 32)
	MOVW	12(R1), R9
	OR		R9, R8
	STXi(8, 7, 0)
	RETURN

/* could be easier if stack was aligned */
TEXT get8(SB), $0				/* ulong -> uvlong */
	MOVW	8(R1), R8
	LDXi(9, 8, 0)
	MOVW	R9, 4(R7)
	SRLXi(9, 9, 32)
	MOVW	R9, 0(R7)
	RETURN

TEXT icflush(SB), $0				/* ulong -> void */
	FLUSHr(7, 0)
	RETURN

TEXT puttba(SB), $0				/* ulong -> void */
	WRPRi(RTBA, 7, 0)
	RETURN

TEXT getpil(SB), $0				/* void -> ulong */
	RDPR(7, RPIL)
	RETURN

/* debugging */
TEXT getpstate(SB), $0			/* void -> ulong */
	RDPR(7, RPSTATE)
	RETURN

/* debugging */
TEXT gettl(SB), $0				/* void -> int */
	RDPR(7, RTL)
	RETURN

/* port: test-and-set */
/* XXX review Ultrasparc User's manual 5.5.1 for all atomic load/stores */
TEXT tas(SB), $0
	TAS (R7), R7
	RETURN

/* port: atomic increment */
TEXT _xinc(SB), $0
  xincagain:
	MOVW	(R7), R8
	ADD		$1, R8, R9
	CASAr(9, 7, 8, ASI_PRIMARY)
	CMP		R8, R9
	BNE		xincagain
	RETURN

/* port: atomic decrement - return new value */
TEXT _xdec(SB), $0
 xdecagain:
	MOVW	(R7), R8
	ADD		$-1, R8, R9
	CASAr(9, 7, 8, ASI_PRIMARY)
	CMP		R8, R9
	BNE		xdecagain
	ADD		$1, R8, R7
	RETURN

/* 
 * port: note: order here is important.  
 * port depends on spllo and spldone addresses.
 */
TEXT spllo(SB), $0
	RDPR(7, RPIL)
	WRPRi(RPIL, 0, 0)
	RETURN

/* port */
TEXT splhi(SB), $0
	MOVW	R15, 4(R(MACH))
	RDPR(7, RPIL)
	WRPRi(RPIL, 0, 15)
	RETURN

/* port */
TEXT splx(SB), $0
	MOVW	R15, 4(R(MACH))
	WRPRi(RPIL, 7, 0)
	RETURN

/* port: end of spl stuff */
TEXT spldone(SB), $0
	RETURN

/* port: are we spllo? */
TEXT islo(SB), $0
	RDPR(8, RPIL)
	MOVW	R0, R7
	MOVRi(7, 8, 1, RCOND_Z)
	RETURN

/* 
 * call openfirmware 
 * Currently only supported prior to mmuinit.
 */
TEXT ofcall(SB), $0
	/* store caller-saveds in local regs */
	MOVW	R1, R17
	MOVW	R2, R18
	MOVW	R3, R19
	MOVW	R4, R20
	MOVW	R5, R21
	MOVW	R6, R22

	/* get sparc64 stack pointer, and perform call with our arg */
	MOVW	R7, R8
	MOVW	ofwentry(SB), R16
	ANDN	$0xf, R1
	SUB		$(BIAS + FRSIZE), R1, R14
	JMPL		(R16)

	/* restore our registers (including SP, no need for fixup) and return */
	MOVW	R17, R1
	MOVW	R18, R2
	MOVW	R19, R3
	MOVW	R20, R4
	MOVW	R21, R5
	MOVW	R22, R6
	MOVW	R8, R7
	RETURN

/* port: given addr of first arg, fetch the return address */
TEXT getcallerpc(SB), $0
	MOVW	0(R1), R7
	RETURN

/* port: ensure memory ordering */
TEXT coherence(SB), $0
	MEMBAR(MBSYNC)
	RETURN

/* port */
TEXT setlabel(SB), $0
	MOVW	R1, (R7)
	MOVW	R15, 4(R7)
	MOVW	$0, R7
	RETURN

/* port */
TEXT gotolabel(SB), $0
	MOVW	(R7), R1
	MOVW	4(R7), R15
	MOVW	$1, R7
	RETURN

/* 
 * XXX we could save two-at-a-time with MOVD, but
 * we would have to enforce alignment of the fpsave region
 * which isn't currently done. 
 */
TEXT savefpregs(SB), $0
	MOVW	FSR, (0*4+32*4)(R7)
	RDST(8, RFPRS)
	MOVW	R8, (1*4+32*4)(R7)
	FMOVF	F0, (0*4)(R7)
	FMOVF	F1, (1*4)(R7)
	FMOVF	F2, (2*4)(R7)
	FMOVF	F3, (3*4)(R7)
	FMOVF	F4, (4*4)(R7)
	FMOVF	F5, (5*4)(R7)
	FMOVF	F6, (6*4)(R7)
	FMOVF	F7, (7*4)(R7)
	FMOVF	F8, (8*4)(R7)
	FMOVF	F9, (9*4)(R7)
	FMOVF	F10, (10*4)(R7)
	FMOVF	F11, (11*4)(R7)
	FMOVF	F12, (12*4)(R7)
	FMOVF	F13, (13*4)(R7)
	FMOVF	F14, (14*4)(R7)
	FMOVF	F15, (15*4)(R7)
	FMOVF	F16, (16*4)(R7)
	FMOVF	F17, (17*4)(R7)
	FMOVF	F18, (18*4)(R7)
	FMOVF	F19, (19*4)(R7)
	FMOVF	F20, (20*4)(R7)
	FMOVF	F21, (21*4)(R7)
	FMOVF	F22, (22*4)(R7)
	FMOVF	F23, (23*4)(R7)
	FMOVF	F24, (24*4)(R7)
	FMOVF	F25, (25*4)(R7)
	FMOVF	F26, (26*4)(R7)
	FMOVF	F27, (27*4)(R7)
	FMOVF	F28, (28*4)(R7)
	FMOVF	F29, (29*4)(R7)
	FMOVF	F30, (30*4)(R7)
	FMOVF	F31, (31*4)(R7)
	RETURN

TEXT restorefpregs(SB), $0
	MOVW	(0*4+32*4)(R7), FSR
	MOVW	(1*4+32*4)(R7), R8
	WRSTi(RFPRS, 8, 0)
	FMOVF	(0*4)(R7), F0
	FMOVF	(1*4)(R7), F1
	FMOVF	(2*4)(R7), F2
	FMOVF	(3*4)(R7), F3
	FMOVF	(4*4)(R7), F4
	FMOVF	(5*4)(R7), F5
	FMOVF	(6*4)(R7), F6
	FMOVF	(7*4)(R7), F7
	FMOVF	(8*4)(R7), F8
	FMOVF	(9*4)(R7), F9
	FMOVF	(10*4)(R7), F10
	FMOVF	(11*4)(R7), F11
	FMOVF	(12*4)(R7), F12
	FMOVF	(13*4)(R7), F13
	FMOVF	(14*4)(R7), F14
	FMOVF	(15*4)(R7), F15
	FMOVF	(16*4)(R7), F16
	FMOVF	(17*4)(R7), F17
	FMOVF	(18*4)(R7), F18
	FMOVF	(19*4)(R7), F19
	FMOVF	(20*4)(R7), F20
	FMOVF	(21*4)(R7), F21
	FMOVF	(22*4)(R7), F22
	FMOVF	(23*4)(R7), F23
	FMOVF	(24*4)(R7), F24
	FMOVF	(25*4)(R7), F25
	FMOVF	(26*4)(R7), F26
	FMOVF	(27*4)(R7), F27
	FMOVF	(28*4)(R7), F28
	FMOVF	(29*4)(R7), F29
	FMOVF	(30*4)(R7), F30
	FMOVF	(31*4)(R7), F31
	RETURN

/* enter userland for the first time with a fake return */
TEXT touser(SB), $0
	WRPRi(RTL, 0, 1)
	MOVW	$(PSUSER<<8), R8
	WRPRi(RTSTATE, 8, 0)
	MOVW	$(UTZERO+32), R8
	WRPRi(RTPC, 8, 0)
	ADD		$4, R8
	WRPRi(RTNPC, 8, 0)
	WRPRi(RPIL, 0, 0)
	MOVW	R7, R1
	RETRY

/* 
 * entry point for all traps except syscall.
 *
 * At entry, the system has incremented TL, saved state in
 * TT/TPC/TNPC/TSTATE, switched to a set of alternate globals.
 * The trap vector then executed:
 *      MOV $1, RCWP
 *      CALL traplink(SB)
 * Which frees up R8..R23 as scratch registers.
 */
TEXT traplink(SB), $-4
	WRPRi(RPSTATE, 0, PSKERNEL)			/* back to normal globals */
	RDPR(8, RTL)
	CMP		R8, $3
	BGE		trapcrit
	RDPR(8, RTT)
/*
	CMP		R8, $0x32
	BE		trapcrit
*/

	RDPR(19, RTSTATE)
	ANDCC	$(PSPRIV<<8), R19, R0
	BE		trapuser

 trapkernel:
	/* use current stack */
	ANDN	$7, R1, R8		/* dbl aligned */
	MOVW	R2, R9
	MOVW	$setSB(SB), R2
	SUB		$UREGSZ, R8
	MOVW	R1, (4*1)(R8)		/* save R1=SP */
	MOVW	R9, (4*2)(R8)		/* SB */
	MOVW	R5, (4*5)(R8)		/* PROC */
	MOVW	R6, (4*6)(R8)		/* MACH */
	MOVW	R8, R1
	JMP		trapcommon

 trapuser:
	/* use mach->proc->kstack + KSTACK as stack */
	MOVW	R2, R10
	MOVW	$setSB(SB), R2
	MOVW	$mach0(SB), R8
	MOVW	(4*2)(R8), R13		/* m->proc */
	MOVW	(4*2)(R13), R9		/* m->proc->kstack */
	ADD		$(KSTACK-UREGSZ), R9
	MOVW	R1, (4*1)(R9)		/* SP */
	MOVW	R10, (4*2)(R9)		/* SB */
	MOVW	R5, (4*5)(R9)		/* PROC */
	MOVW	R6, (4*6)(R9)		/* MACH */
	MOVW	R9, R1
	MOVW	R13, R5
	MOVW	R8, R6

 trapcommon:
	/* R1, R2, R5 and R6 already taken care of */
	MOVW	R3, (4*3)(R1)
	MOVW	R4, (4*4)(R1)
	MOVW	R7, (4*7)(R1)
	WRPRi(RCWP, 0, 0)				/* get back R8..R23 */
	MOVW	R8, (4*8)(R1)
	MOVW	R9, (4*9)(R1)
	MOVW	R10, (4*10)(R1)
	MOVW	R11, (4*11)(R1)
	MOVW	R12, (4*12)(R1)
	MOVW	R13, (4*13)(R1)
	MOVW	R14, (4*14)(R1)
	MOVW	R15, (4*15)(R1)
	MOVW	R16, (4*16)(R1)
	MOVW	R17, (4*17)(R1)
	MOVW	R18, (4*18)(R1)
	MOVW	R19, (4*19)(R1)
	MOVW	R20, (4*20)(R1)
	MOVW	R21, (4*21)(R1)
	MOVW	R22, (4*22)(R1)
	MOVW	R23, (4*23)(R1)
	MOVW	R24, (4*24)(R1)
	MOVW	R25, (4*25)(R1)
	MOVW	R26, (4*26)(R1)
	MOVW	R27, (4*27)(R1)
	MOVW	R28, (4*28)(R1)
	MOVW	R29, (4*29)(R1)
	MOVW	R30, (4*30)(R1)
	MOVW	R31, (4*31)(R1)
	MOVW	Y, R8
	MOVW	R8, (4*(32+0))(R1)		/* Y */
	RDPR(8, RTT)
	MOVW	R8, (4*(32+1))(R1)
	RDPR(8, RTSTATE)
	SRLXi(8, 8, 8)
	MOVW	R8, (4*(32+2))(R1)		/* pstate (and ASI/CCR) */
	RDPR(8, RTNPC)
	MOVW	R8, (4*(32+3))(R1)		/* nPC */
	RDPR(8, RTPC)
	MOVW	R8, (4*(32+4))(R1)		/* PC */
	WRPRi(RTL, 0, 0)

	MOVW	R1, R7				/* pointer to Ureg */
	SUB		$8, R1
	JMPL	trap(SB)
	ADD	$8, R1

 traprestore:
	/* restore all registers */
	MOVW	(4*31)(R1), R31
	MOVW	(4*30)(R1), R30
	MOVW	(4*29)(R1), R29
	MOVW	(4*28)(R1), R28
	MOVW	(4*27)(R1), R27
	MOVW	(4*26)(R1), R26
	MOVW	(4*25)(R1), R25
	MOVW	(4*24)(R1), R24
	MOVW	(4*23)(R1), R23
	MOVW	(4*22)(R1), R22
	MOVW	(4*21)(R1), R21
	MOVW	(4*20)(R1), R20
	MOVW	(4*19)(R1), R19
	MOVW	(4*18)(R1), R18
	MOVW	(4*17)(R1), R17
	MOVW	(4*16)(R1), R16
	MOVW	(4*15)(R1), R15
	MOVW	(4*14)(R1), R14
	MOVW	(4*13)(R1), R13
	MOVW	(4*12)(R1), R12
	MOVW	(4*11)(R1), R11
	MOVW	(4*10)(R1), R10
	MOVW	(4*9)(R1), R9
	MOVW	(4*8)(R1), R8
	MOVW	(4*7)(R1), R7

 sysrestore:
	/* restore all but caller-saved registers and retval */
	WRPRi(RPSTATE, 0, PSKERNEL)
	WRPRi(RTL, 0, 1)
	MOVW	(4*(32+4))(R1), R3		/* PC */
	WRPRi(RTPC, 3, 0)
	MOVW	(4*(32+3))(R1), R3		/* nPC */
	WRPRi(RTNPC, 3, 0)
	MOVW	(4*(32+2))(R1), R3		/* pstate (and ASI/CCR) */
	SLLXi(3, 3, 8)
	WRPRi(RTSTATE, 3, 0)
	MOVW	(4*(32+0))(R1), R3		/* Y */
	MOVW	R3, Y
	MOVW	(4*15)(R1), R15
	MOVW	(4*6)(R1), R6
	MOVW	(4*5)(R1), R5
	MOVW	(4*4)(R1), R4
	MOVW	(4*3)(R1), R3
	MOVW	(4*2)(R1), R2
	MOVW	(4*1)(R1), R1			/* restore R1=SP */
	RETRY

 trapcrit:
	/* get off one final warning */
	MOVW	R1, R7
	MOVW	$setSB(SB), R2
	MOVW	$(BOOTSTACK+KSTACK-16), R1
	RDPR(8, RTPC)
	MOVW	R8, 8(R1)
	RDPR(8, RTT)
	MOVW	R8, 12(R1)
	JMPL		trapcritical(SB)
	JMPL		exit(SB)

/*
 * entrypoint for syscall trap.
 *
 * At entry, the system has incremented TL, saved state in
 * TT/TPC/TNPC/TSTATE, switched to a set of alternate globals.
 * The trap vector then executed:
 *      MOV $1, RCWP
 *      CALL traplink(SB)
 * Which frees up R8..R23 as scratch registers.
 */
TEXT syslink(SB), $-4
	/* 
	 * we assume we're coming from userland -- syscall() will check.
	 * get back our globals and use mach->proc->kstack+KSTACK.
	 * We only need to preserve R1..R7 and R15.
	 */
	WRPRi(RCWP, 0, 0)
	WRPRi(RPSTATE, 0, PSKERNEL)
	MOVW	R2, R10
	MOVW	$setSB(SB), R2
	MOVW	$mach0(SB), R8
	MOVW	(4*2)(R8), R13		/* m->proc */
	MOVW	(4*2)(R13), R9		/* m->proc->kstack */
	ADD		$(KSTACK-UREGSZ), R9

	MOVW	R1, (4*1)(R9)		/* SP */
	MOVW	R10, (4*2)(R9)		/* R2 */
	MOVW	R3, (4*3)(R9)
	MOVW	R4, (4*4)(R9)
	MOVW	R5, (4*5)(R9)
	MOVW	R6, (4*6)(R9)
	MOVW	R7, (4*7)(R9)
	MOVW	R15, (4*15)(R9)		/* return addr */
	MOVW	R9, R1
	MOVW	R13, R5
	MOVW	R8, R6

	MOVW	Y, R8
	MOVW	R8, (4*(32+0))(R1)		/* Y */
	RDPR(8, RTSTATE)
	SRLXi(8, 8, 8)
	MOVW	R8, (4*(32+2))(R1)		/* pstate (and ASI/CCR) */
	RDPR(8, RTNPC)
	MOVW	R8, (4*(32+3))(R1)		/* nPC */
	RDPR(8, RTPC)
	MOVW	R8, (4*(32+4))(R1)		/* PC */

	RDPR(8, RPIL)					/* XXX PIL into pad for now */
	MOVW	R8, (4*(32+5))(R1)

	MOVW	R1, R7				/* pointer to Ureg */
	SUB		$8, R1
	WRPRi(RTL, 0, 0)
	WRPRi(RPSTATE, 0, PSKERNEL|PSIE)
	JMPL		syscall(SB)
	ADD		$8, R1
	JMP		sysrestore

/* 
 * defined in port but only used in l.s/trap.c.
 * fake a return from fork syscall with returnval 0.
 */
TEXT forkret(SB), $0
	WRPRi(RPIL, 0, 0)
	MOVW	R0, R7
	ADD		$8, R1
	JMP		sysrestore

TEXT intrdisable(SB), $0
	WRPRi(RPSTATE, 0, PSKERNEL)
	RETURN

TEXT intrenable(SB), $0
	WRPRi(RPSTATE, 0, PSKERNEL|PSIE)
	RETURN

TEXT intrraise(SB), $0
	WRSTi(RSETINT, 7, 0)
	RETURN

TEXT intrclear(SB), $0
	WRSTi(RCLRINT, 7, 0)
	RETURN

/* debug */
TEXT getsoftint(SB), $0			/* void -> ulong */
	RDST(7, RSOFTINT)
	RETURN

TEXT reset(SB), $0
	SHUTDOWN
  rloop:
	JMP		rloop
	RETURN

