/*
 * Memory and machine-specific definitions.  Used in C and assembler.
 */

/*
 * Sizes
 */
#define	BY2WD		4			/* port: bytes per word */
#define	BY2V			8			/* port: bytes per vlong */
#define	BY2PG		8192		/* port: bytes per page */
#define	ROUND(s, sz)	(((s)+((sz)-1))&~((sz)-1))	/* port */
#define	PGSHIFT		13			/* port: log(BY2PG) */
#define	PGROUND(s)	ROUND(s, BY2PG)	/* port */
#define	BLOCKALIGN	8			/* port */
#define	KSTACK		BY2PG		/* port: Size of kernel stack */

/*
 * Time
 */
#define	HZ		(100)			/* port: clock frequency */
#define	TK2SEC(t)	((t)/HZ)			/* port: ticks to seconds */

/*
 * Address space.
 *
 * User is at 0-2GB.
 * Kernel is at 2GB-4GB.
 */
#define	KZERO		0x80000000	/* port: start of kernel memory */
#define	TRAPS		KZERO		/* trap table */
#define	BOOTSTACK	(TRAPS+0x8000)	/* boot stack */
#define	KTZERO		(BOOTSTACK+KSTACK)	/* port: start of kernel text */
#define	UTZERO		BY2PG		/* port: first address in user text */
#define	TSTKTOP		KZERO		/* port: top of temporary stack */
#define 	TSTKSIZ 		50			/* port: pages for TSTK */
#define	USTKTOP		(TSTKTOP-TSTKSIZ*BY2PG)	/* port: top of user stack */
#define	USTKSIZE		(4*MB)		/* port: size of user stack */

#define	PGMAPSIZE	(32*MB)		/* memory space for non-linear mapped pages - multiple of 4MB */
#define	NPGMAP		(PGMAPSIZE/BY2PG)
#define	MAPSEGBASE	(0-PGMAPSIZE)	/* VA for non-linear mappings */

#define	RAMTOP		(MAPSEGBASE - KZERO)	/* highest PA for ram */

/*
 * MMU api and hardware definitions.
 * note: right now our ulong is 32bits, while pte's are 64.
 * We borrow a software bit to represent the hardware valid bit 
 * (which is the only bit above 32 that the port code needs).
 */
#define	PTEVALID		(1<<7)		/* software */
#define	PTEVALIDHW	(1LL<<63)
#define	PTEGLOB		(1<<0)
#define	PTERONLY	(0<<1)
#define	PTEWRITE		(1<<1)
#define	PTEKERNEL	(1<<2)
#define	PTEIO		(1<<3)
#define	PTEUNCACHED	(1<<8)		/* software */
#define	PTECACHE	(3<<4)
#define	PTEPHYCACHE	(2<<4)
#define	PTELOCK		(1<<6)
#define	PTE8KB		(0L<<61)
#define		PTE64KB		(1LL<<61)
#define		PTE512KB		(2LL<<61)
#define		PTE4MB		(3LL<<61)	
#define	PPN(pa)		((pa)&0xFFFFE000)	/* port */

#define	PTEMAPMEM	(1024*1024)			/* port */
#define	PTEPERTAB	(PTEMAPMEM/BY2PG)	/* port */
#define 	SEGMAPSIZE	1984				/* port */
#define	SSEGMAPSIZE	16					/* port */

/* 8K pages, 16K alias boundary */
#define	NCOLOR	2						/* port */
#define 	getpgcolor(va)	(((va) >> PGSHIFT) & (NCOLOR-1))	/* port */

/*
 * MMU access
 */
#define ASI_MMU_I			0x50
#define	MMU_CTX_PRI			0x8		/* shared - DMMU only */
#define	MMU_CTX_SEC			0x10		/* shared - DMMU only */
#define	MMU_FAULTSTATUS		0x18
#define	MMU_FAULTADDR		0x20		/* DMMU only */
#define	MMU_TAGACCESS		0x30
#define ASI_MMU_IDATAIN	0x54
#define ASI_MMU_IDATA		0x55
#define ASI_MMU_ITAG		0x56
#define ASI_MMU_IDEMAP	0x57
#define	DEMAPCTX			(1<<6)
#define ASI_MMU_D			0x58
#define ASI_MMU_DDATAIN	0x5c
#define ASI_MMU_DDATA		0x5d
#define ASI_MMU_DTAG		0x5e
#define ASI_MMU_DDEMAP	0x5f

#define ASI_LSU_CONTROL	0x45
#define ASI_INTR_RECEIVE	0x49
#define ASI_AFSR			0x4c
#define ASI_AFAR			0x4d
#define ASI_UDB			0x7f
#define	UDB_DATA0			0x40
#define ASI_PRIMARY		0x80

/* FSR bits */
#define FSR_FV	(1<<0)
#define FSR_OW	(1<<1)
#define FSR_W		(1<<2)
#define FSR_PR	(1<<3)
#define FSR_E		(1<<6)

/*
 * PSTATE bits
 */
#define PSAG		(1<<0)
#define PSIE		(1<<1)
#define PSPRIV		(1<<2)
#define PSAM		(1<<3)
#define PSPEF		(1<<4)
#define PSRED		(1<<5)
#define PSTSO		(0<<6)
#define 	PSPSO		(1<<6)
#define 	PSRMO		(2<<6)
#define PSTLE		(1<<8)
#define PSCLE		(1<<9)

/*
 * Magic registers
 */
#define	MACH		6		/* R6 is m-> */
#define	PROC		5		/* R5 is up-> */
