
#include <u.h>
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"

static char *logbuf = (char*)0x2000;
static int logpos = 0;
enum { TxReady = 1<<2 };
typedef struct {
	struct {
		uchar	csr;
		uchar	dummy1;
		uchar	dr;
		uchar	dummy2;
	} *ptr;
} Ctrl;
static Ctrl dev = { (void*)0xfffa6004 };

static uchar
csr8r(Ctrl *cp, int addr)
{
	cp->ptr->csr = addr;
	return cp->ptr->csr;
}

void
zsetdev(ulong va)
{
	dev.ptr = (void*)va;
}

static void
zputc(int c)
{
#ifndef QUIET
	if(dev.ptr) {
		while(!(csr8r(&dev, 0) & TxReady))
			continue;
		dev.ptr->dr = c;
	}
#endif
}

static void
logputc(int c)
{
	logbuf[logpos] = c;
	logpos = (logpos + 1) & 0xfff;
}


/* pick which we want to use */
void (*xputcfunc)(int) = zputc;
int xputcCR = 1;

static void
printint(uint val, int base)
{
	static char *valch = "0123456789abcdef";
	int x, b;

	for(b = 1; val / b >= base; b = b * base)
		continue;
	while(b) {
		x = val / b;
		val -= x * b;
		b /= base;
		xputcfunc(valch[x]);
	}
}

void
xputs(char *str, int n)
{
	while(n > 0) {
		if(*str == '\n' && xputcCR)
			xputcfunc('\r');
		xputcfunc(*str++);
		n--;
	}
}

void
xwrite(void *buf, int n)
{
    char *b = (char*)buf;
    while(n--)
        xputcfunc(*b++);
}

void
xprint(char *fmt, ...)
{
	char *str, ch;
	va_list arg;
	int s;

	s = splhi();
	va_start(arg, fmt);
	while(*fmt) {
		switch(*fmt) {
		case '%':
			fmt++;
			switch(*fmt) {
			case '\0':
				fmt--;
				break;
			case 'c':
				ch = va_arg(arg, int);
				xputcfunc(ch);
				break;
			case 'd':
				printint(va_arg(arg, int), 10);
				break;
			case 's':
				str = va_arg(arg, char*);
				xputs(str, strlen(str));
				break;
			case 'x':
				printint(va_arg(arg, int), 16);
				break;
			default:
				xputcfunc(*fmt);
				break;
			}
			break;
		case '\n':
			if(xputcCR)
				xputcfunc('\r');
			xputcfunc('\n');
			break;
		default:
			xputcfunc(*fmt);
			break;
		}
		fmt++;
	}
	va_end(arg);
	splx(s);
}


